/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#include "../mimimi.h"

static unsigned char mimimi_layer_pixel(struct mimimi_layer *layer, unsigned char y_angle, int x, int y)
{
	int w, h;
	int w0, w1;
	int s;
	int count;
	int a;
	
	h = layer->count;
	
	if (y < 0 || y >= h) return 0;
	
	w0 = layer->rows[y].width;
	w1 = layer->rows[y].height;
	
	s = 128 - mimimi_cosine[y_angle * 16 % 2048];
	w = s * w0 + (256 - s) * w1;
	w /= 512;
	
	if (x < -w || x >= w) return 0;
	
	count = layer->rows[y].count;
	
	a = y_angle + 64;
	a *= count;
	a += 128;
	a /= 256;
	a += x;
	a = mimimi_mod(a, count);
	
	return layer->rows[y].colors[a];
}

static void mimimi_swap(int *a, int *b)
{
	int c;
	c = *a;
	*a = *b;
	*b = c;
}

static void mimimi_layer(struct mimimi_image *image, struct mimimi_layer *layer, int x0, int y0, struct mimimi_pose_layer *pose_layer, int behind)
{
	int width, height;
	unsigned char y_angle;
	int z_angle;
	int x, y;
	int x1, y1;
	unsigned char color;
	
	width = image->width;
	height = image->height;
	
	y_angle = pose_layer->y_angle;
	z_angle = mimimi_mod(pose_layer->z_angle, 4);
	
	if (behind != 0) y_angle += 128;
	
	if (z_angle == 0) {
		
		for (y = 0 ; y < height ; y++) {
			
			for (x = 0 ; x < width ; x++) {
				
				x1 = x + layer->x - x0;
				y1 = y + layer->y - y0;
				
				x1 += pose_layer->slant * (y - y0) / 256;
				
				if (behind != 0) x1 = -x1 - 1;
				
				color = mimimi_layer_pixel(layer, y_angle, x1, y1);
				if (color != 0) image->colors[x + y * width] = color;
			}
		}
	}
	
	if (z_angle == 1) {
		
		for (y = 0 ; y < height ; y++) {
			
			for (x = 0 ; x < width ; x++) {
				
				x1 = x - layer->y - x0;
				y1 = y + layer->x - y0;
				
				y1 -= pose_layer->slant * (x - x0) / 256;
				
				if (behind != 0) y1 = -y1 - 1;
				
				color = mimimi_layer_pixel(layer, y_angle, y1, -x1);
				if (color != 0) image->colors[x + y * width] = color;
			}
		}
	}
	
	if (z_angle == 2) {
		
		for (y = 0 ; y < height ; y++) {
			
			for (x = 0 ; x < width ; x++) {
				
				x1 = x - layer->x - x0;
				y1 = y - layer->y - y0;
				
				x1 += pose_layer->slant * (y - y0) / 256;
				
				if (behind != 0) x1 = -x1 + 1;
				
				color = mimimi_layer_pixel(layer, y_angle, -x1, -y1);
				if (color != 0) image->colors[x + y * width] = color;
			}
		}
	}
	
	if (z_angle == 3) {
		
		for (y = 0 ; y < height ; y++) {
			
			for (x = 0 ; x < width ; x++) {
				
				x1 = x + layer->y - x0;
				y1 = y - layer->x - y0;
				
				y1 -= pose_layer->slant * (x - x0) / 256;
				
				if (behind != 0) y1 = -y1 + 1;
				
				color = mimimi_layer_pixel(layer, y_angle, -y1, x1);
				if (color != 0) image->colors[x + y * width] = color;
			}
		}
	}
}

static void mimimi_apply_y_angle(struct mimimi_model *model, struct mimimi_layer *layers, struct mimimi_pose_layer *pose_layers)
{
	int i, parent_index;
	struct mimimi_layer *layer;
	struct mimimi_pose_layer *pose_layer;
	int parent_y_angle;
	unsigned char y_angle;
	int sin, cos;
	int x, z;
	
	for (i = 0 ; i < model->count ; i++) {
		
		layer = layers + i;
		pose_layer = pose_layers + i;
		parent_index = model->layers[i].parent_index;
		
		if (i == parent_index) continue;
		
		parent_y_angle = pose_layers[parent_index].y_angle;
		pose_layer->y_angle += parent_y_angle;
		
		y_angle = parent_y_angle;
		
		sin = mimimi_sine[y_angle * 8];
		cos = mimimi_cosine[y_angle * 8];
		
		x = layer->x;
		z = layer->z;
		
		layer->x = cos * x / 128 + sin * z / 128;
		layer->z = cos * z / 128 - sin * x / 128;
	}
}

static void mimimi_apply_z_angle(struct mimimi_model *model, struct mimimi_layer *layers, struct mimimi_pose_layer *pose_layers)
{
	static int xs[0x80], ys[0x80];
	
	int i;
	struct mimimi_layer *layer;
	struct mimimi_pose_layer *pose_layer;
	int parent_index;
	int z_angle;
	int x, y;
	
	for (i = 0 ; i < model->count ; i++) {
		
		layer = layers + i;
		pose_layer = pose_layers + i;
		parent_index = model->layers[i].parent_index;
		
		xs[i] = layer->x;
		ys[i] = layer->y;
		if (i == parent_index) continue;
		xs[i] += xs[parent_index];
		ys[i] += ys[parent_index];
		
		z_angle = mimimi_mod(pose_layer->z_angle, 4);
		
		x = xs[i];
		y = ys[i];
		if (z_angle == 1) {
			x = -x;
			mimimi_swap(&x, &y);
		}
		if (z_angle == 2) {
			x = -x;
			y = -y;
		}
		if (z_angle == 3) {
			y = -y;
			mimimi_swap(&x, &y);
		}
		layer->x = x - xs[parent_index];
		layer->y = y - ys[parent_index];
		
		if (z_angle == 1) {
			xs[i] = -xs[i];
			mimimi_swap(xs + i, ys + i);
		}
		if (z_angle == 2) {
			xs[i] = -xs[i];
			ys[i] = -ys[i];
		}
		if (z_angle == 3) {
			ys[i] = -ys[i];
			mimimi_swap(xs + i, ys + i);
		}
	}
}

static void mimimi_apply_slant(struct mimimi_model *model, struct mimimi_layer *layers, struct mimimi_pose_layer *pose_layers)
{
	static int ys[0x80];
	
	int i;
	struct mimimi_layer *layer;
	struct mimimi_pose_layer *pose_layer;
	int parent_index;
	
	for (i = 0 ; i < model->count ; i++) {
		
		layer = layers + i;
		pose_layer = pose_layers + i;
		parent_index = model->layers[i].parent_index;
		
		ys[i] = layer->y + pose_layer->y;
		if (i == parent_index) continue;
		ys[i] += ys[parent_index];
		
		layer->x += pose_layer->slant * ys[i] / 256;
	}
}

static void mimimi_absolutise(struct mimimi_model *model, struct mimimi_layer *layers, struct mimimi_pose_layer *pose_layers)
{
	int i;
	struct mimimi_layer *layer;
	struct mimimi_pose_layer *pose_layer;
	int parent_index;
	
	for (i = 0 ; i < model->count ; i++) {
		
		layer = layers + i;
		pose_layer = pose_layers + i;
		parent_index = model->layers[i].parent_index;
		
		layer->y += pose_layer->y;
		
		if (i == parent_index) continue;
		
		layer->x += layers[parent_index].x;
		layer->y += layers[parent_index].y;
		layer->z += layers[parent_index].z;
		layer->amplifier = layer->amplifier * layers[parent_index].amplifier / 32;
		
		pose_layer->slant += pose_layers[parent_index].slant;
		pose_layer->z_angle += pose_layers[parent_index].z_angle;
		pose_layer->z_angle = mimimi_mod(pose_layer->z_angle, 4);
	}
}

struct mimimi_half_layer {
	int behind;
	int z;
	struct mimimi_layer *layer;
	struct mimimi_pose_layer *pose_layer;
};

void mimimi_pose(struct mimimi_image *image, struct mimimi_model *model, struct mimimi_pose *pose)
{
	static struct mimimi_layer layers[0x80];
	static struct mimimi_pose_layer pose_layers[0x80];
	static struct mimimi_half_layer half_layers[0x100];
	
	struct mimimi_layer *layer;
	struct mimimi_pose_layer *pose_layer;
	struct mimimi_half_layer *half_layer, half_layer0;
	int i, j, k;
	int a, b;
	
	for (i = 0 ; i < model->count ; i++) {
		layers[i] = *model->layers[i].layer;
		pose_layers[i] = pose->layers[i];
	}
	
	mimimi_apply_y_angle(model, layers, pose_layers);
	mimimi_apply_z_angle(model, layers, pose_layers);
	mimimi_apply_slant(model, layers, pose_layers);
	mimimi_absolutise(model, layers, pose_layers);
	
	for (i = 0 ; i < model->count ; i++) {
		
		layer = layers + i;
		pose_layer = pose_layers + i;
		
		half_layers[i * 2 + 0].behind = 0;
		half_layers[i * 2 + 0].z = layer->z * layer->amplifier + layer->width * 32;
		half_layers[i * 2 + 0].layer = layer;
		half_layers[i * 2 + 0].pose_layer = pose_layer;
		
		half_layers[i * 2 + 1].behind = 1;
		half_layers[i * 2 + 1].z = layer->z * layer->amplifier - layer->width * 32;
		half_layers[i * 2 + 1].layer = layer;
		half_layers[i * 2 + 1].pose_layer = pose_layer;
	}
	
	/* z ordering */
	for (i = 0 ; i < model->count * 2 ; i++) {
		
		a = half_layers[i].z;
		
		k = i;
		for (j = i - 1 ; j >= 0 ; j--) {
			
			b = half_layers[j].z;
			if (a >= b) break;
			
			half_layer0 = half_layers[k];
			half_layers[k] = half_layers[j];
			half_layers[j] = half_layer0;
			
			k = j;
		}
	}
	
	for (i = 0 ; i < model->count * 2 ; i++) {
		half_layer = half_layers + i;
		mimimi_layer(image, half_layer->layer, pose->x, pose->y, half_layer->pose_layer, half_layer->behind);
	}
}

static void mimimi_interpolate_layer(struct mimimi_pose_layer *a, struct mimimi_pose_layer *b, int i, int count)
{
	int y, slant, y_angle, z_angle;
	
	y = a->y * (count - i);
	y += b->y * i;
	
	slant = a->slant * (count - i);
	slant += b->slant * i;
	
	y_angle = a->y_angle * (count - i);
	y_angle += b->y_angle * i;
	
	z_angle = a->z_angle * (count - i);
	z_angle += b->z_angle * i;
	
	a->y = mimimi_div(y, count);
	a->slant = mimimi_div(slant, count);
	a->y_angle = mimimi_div(y_angle, count);
	a->z_angle = mimimi_div(z_angle, count);
}

void mimimi_interpolate(struct mimimi_pose *a, struct mimimi_pose *b, int i, int count)
{
	int x, y;
	int j;
	
	i = 128 - mimimi_cosine[i * 1024 / count];
	count = 256;
	
	x = a->x * (count - i);
	x += b->x * i;
	
	y = a->y * (count - i);
	y += b->y * i;
	
	a->x = mimimi_div(x, count);
	a->y = mimimi_div(y, count);
	
	for (j = 0 ; j < (int) (sizeof a->layers / sizeof *a->layers) ; j++) mimimi_interpolate_layer(a->layers + j, b->layers + j, i, count);
}
