/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#include <stdlib.h>
#include "../mimimi.h"

static int mimimi_wall_tile(struct mimimi_ground *ground, int x, int y)
{
	if (mimimi_ground_tile(ground, x, y + 1) == 0 && mimimi_ground_tile(ground, x, y - 1) == 0) return 0;
	return mimimi_ground_tile(ground, x, y);
}

static void mimimi_wall_physics(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	int x0, y0;
	int x1, y1;
	int left, right;
	int top;
	
	x0 = position->x;
	y0 = position->y;
	x1 = physics->width / 2;
	y1 = physics->height;
	
	top = mimimi_div_down(y0 - y1, 128);
	left = mimimi_div_down(x0 - x1, 128);
	right = mimimi_div_down(x0 + x1, 128);
	
	if (mimimi_wall_tile(ground, left, top) != 0) {
		if (physics->dx < 0) physics->dx = 0;
		position->x = (left + 1) * 128 + x1;
	}
	
	if (mimimi_wall_tile(ground, right, top) != 0) {
		if (physics->dx > 0) physics->dx = 0;
		position->x = right * 128 - x1 - 1;
	}
}

static void mimimi_slope_physics(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	int x0, y0;
	int x1, y1;
	int top, bottom;
	int left, right;
	
	x0 = position->x;
	y0 = position->y;
	x1 = physics->width / 2;
	y1 = physics->height;
	
	top = mimimi_div_down(y0 - y1, 128);
	bottom = mimimi_div_down(y0 - 1, 128);
	left = mimimi_div_down(x0 - x1, 128);
	right = mimimi_div_down(x0 + x1, 128);
	
	if (mimimi_ground_tile(ground, left, bottom) != 0) {
		if (mimimi_ground_tile(ground, left, top) == 0) {
			position->y = bottom * 128;
			if (physics->sprite != NULL) physics->sprite->trail = 0xFF;
		}
	}
	
	if (mimimi_ground_tile(ground, right, bottom) != 0) {
		if (mimimi_ground_tile(ground, right, top) == 0) {
			position->y = bottom * 128;
			if (physics->sprite != NULL) physics->sprite->trail = 0xFF;
		}
	}
}

static void mimimi_ceiling_physics(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	int x0, y0;
	int x1, y1;
	int top;
	int left, right;
	
	x0 = position->x;
	y0 = position->y;
	x1 = physics->width / 2;
	y1 = physics->height;
	
	top = mimimi_div_down(y0 - y1, 128);
	left = mimimi_div_down(x0 - x1, 128);
	right = mimimi_div_down(x0 + x1, 128);
	
	if (mimimi_wall_tile(ground, left, top) != 0 && mimimi_wall_tile(ground, right, top) != 0) {
		if (physics->dy < 0) physics->dy = 0;
		position->y = (top + 1) * 128 + y1;
	}
}

static void mimimi_landing_physics(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	int x0, y0;
	int x1, y1;
	int left, right;
	int bottom, bottom1;
	unsigned char tile1, tile2;
	int restitution;
	
	if (physics->dy < 0) return;
	
	y0 = position->y - mimimi_div(physics->dy, 256) - 1;
	x0 = position->x;
	x1 = physics->width / 2;
	y1 = position->y;
	
	bottom = mimimi_div_down(y0, 128);
	bottom1 = mimimi_div_down(y1, 128);
	left = mimimi_div_down(x0 - x1, 128);
	right = mimimi_div_down(x0 + x1, 128);
	
	while (bottom <= bottom1) {
		
		tile1 = mimimi_ground_tile(ground, left, bottom);
		tile2 = mimimi_ground_tile(ground, right, bottom);
		if (tile1 == 0 && tile2 == 0) {
			bottom++;
			continue;
		}
		
		if (ground->restitution[tile1] < ground->restitution[tile2]) {
			restitution = ground->restitution[tile1];
		}
		else {
			restitution = ground->restitution[tile2];
		}
		
		position->y = bottom * 128;
		physics->dy = -mimimi_div(physics->dy * restitution, 256);
		if (physics->dy > -2048) {
			physics->airborne = 0;
			physics->dy = 0;
		}
		
		break;
	}
}

static void mimimi_fall_physics(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	int x0, y0;
	int x1;
	int bottom, bottom2;
	int center;
	int left, right;
	
	x0 = position->x;
	y0 = position->y;
	x1 = physics->width / 2;
	
	bottom = mimimi_div_down(y0, 128);
	bottom2 = bottom + 1;
	center = mimimi_div_down(x0, 128);
	left = mimimi_div_down(x0 - x1, 128);
	right = mimimi_div_down(x0 + x1, 128);
	
	if (mimimi_ground_tile(ground, left, bottom) == 0 && mimimi_ground_tile(ground, right, bottom) == 0) {
		if (mimimi_ground_tile(ground, center, bottom2) == 0) {
			physics->airborne = 1;
		}
		else {
			if (physics->sprite != NULL) physics->sprite->trail = 0xFF;
			position->y = bottom2 * 128;
		}
	}
}

static void mimimi_physics_dynamics(struct mimimi_physics *physics, struct mimimi_position *position)
{
	position->x += mimimi_div(physics->dx, 256);
	position->y += mimimi_div(physics->dy, 256);
}

static void mimimi_ground_collision_physics(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	mimimi_slope_physics(physics, position, ground);
	mimimi_wall_physics(physics, position, ground);
	mimimi_fall_physics(physics, position, ground);
	if (physics->airborne == 0) position->y = mimimi_div_down(position->y, 128) * 128;
}

static void mimimi_airborne_collision_physics(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	mimimi_ceiling_physics(physics, position, ground);
	mimimi_wall_physics(physics, position, ground);
	mimimi_landing_physics(physics, position, ground);
}

static void mimimi_ground_physics(struct mimimi_physics *physics)
{
	physics->dx -= mimimi_div(physics->dx, 5);
	physics->dy = 0;
}

static void mimimi_airborne_physics(struct mimimi_physics *physics)
{
	physics->dy += physics->gravity;
	physics->dx -= mimimi_div(physics->dx, 7);
	physics->dy -= mimimi_div(physics->dy, 32);
}

void mimimi_collision_physics_tick(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	if (physics->airborne == 0) mimimi_ground_collision_physics(physics, position, ground);
	else mimimi_airborne_collision_physics(physics, position, ground);
}

void mimimi_physics_tick(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground)
{
	if (physics->airborne == 0) mimimi_ground_physics(physics);
	else mimimi_airborne_physics(physics);
	mimimi_physics_dynamics(physics, position);
	mimimi_collision_physics_tick(physics, position, ground);
}

void mimimi_physics(struct mimimi_physics *physics, int width, int height)
{
	physics->dx = 0;
	physics->dy = 0;
	physics->airborne = 1;
	physics->width = width;
	physics->height = height;
	physics->gravity = 2048;
	physics->sprite = NULL;
}

static void mimimi_walk_tick(struct mimimi_walk *walk, struct mimimi_physics *physics)
{
	int ax;
	
	if (walk->direction == 0) return;
	
	if (physics->airborne == 0) ax = walk->ground_speed;
	else ax = walk->airborne_speed;
	
	if (walk->direction == 1) physics->dx -= ax;
	if (walk->direction == 2) physics->dx += ax;
}

void mimimi_sprite(struct mimimi_sprite *sprite, struct mimimi_model *model, struct mimimi_ground *ground, int x, int y, int width, int height)
{
	sprite->position.x = x;
	sprite->position.y = y;
	
	sprite->trailing_y = y;
	sprite->trail = 0;
	
	sprite->ground = ground;
	mimimi_physics(&sprite->physics, width, height);
	sprite->physics.sprite = sprite;
	
	sprite->walk.ground_speed = 4096;
	sprite->walk.airborne_speed = 2048;
	sprite->walk.direction = 0;
	
	sprite->offset.x = 0;
	sprite->offset.y = 0;
	
	sprite->animation_time = 0;
	sprite->landing_time = 0;
	sprite->direction = 1;
	sprite->model = model;
}

void mimimi_sprite_tick(struct mimimi_sprite *sprite)
{
	mimimi_walk_tick(&sprite->walk, &sprite->physics);
	sprite->position.x -= sprite->offset.x;
	sprite->position.y -= sprite->offset.y;
	mimimi_physics_tick(&sprite->physics, &sprite->position, sprite->ground);
	sprite->position.x += sprite->offset.x;
	sprite->position.y += sprite->offset.y;
}

void mimimi_jump(void *data)
{
	struct mimimi_sprite *sprite;
	sprite = data;
	if (sprite->physics.airborne != 0) return;
	sprite->physics.airborne = 1;
	sprite->physics.dy = -0x6800;
}
