/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#include <stdlib.h>
#include "../mimimi.h"

static int mimimi_ground_floor(struct mimimi_ground *ground, int x)
{
	int y;
	for (y = ground->height ; y > 0 ; y--) {
		if (mimimi_ground_tile(ground, x, y - 1) == 0) return y;
	}
	return 0;
}

static void mimimi_origin(struct mimimi_ground *ground, struct mimimi_position *position)
{
	for (position->y = 0 ; position->y < ground->height ; position->y++) {
		for (position->x = 0 ; position->x < ground->width ; position->x++) {
			if (ground->tiles[position->x + position->y * ground->width] == 2) return;
		}
	}
}

void mimimi_stage(struct mimimi_stage *stage, struct mimimi_ground **grounds, int count)
{
	int i, k;
	struct mimimi_position offset;
	int left_floor_y, right_floor_y;
	struct mimimi_area *area;
	
	stage->sprite_count = 0;
	stage->area_count = count;
	stage->areas = malloc(sizeof *stage->areas * count);
	if (stage->areas == NULL) exit(1);
	stage->clamped_camera.input = &stage->camera;
	stage->platform_set_count = 0;
	
	for (i = 0 ; i < count ; i++) {
		area = stage->areas + i;
		area->ground = grounds[i];
		area->background_count = 0;
		area->foreground_count = 0;
	}
	
	mimimi_origin(grounds[0], &offset);
	offset.x *= -1;
	offset.y *= -1;
	stage->areas[0].ground = grounds[0];
	stage->areas[0].offset = offset;
	
	left_floor_y = mimimi_ground_floor(grounds[0], 0);
	
	for (i = 1 ; i < count ; i++) {
		area = stage->areas + i;
		area->ground = grounds[i];
		right_floor_y = mimimi_ground_floor(grounds[i], grounds[i]->width - 1);
		offset.x -= grounds[i]->width;
		offset.y -= right_floor_y - left_floor_y;
		left_floor_y = mimimi_ground_floor(grounds[i], 0);
		area->offset = offset;
	}
	
	k = sizeof stage->sounds / sizeof *stage->sounds;
	
	for (i = 0 ; i < k ; i++) {
		stage->sounds[i].sound.count = 0;
		stage->sounds[i].current = 0;
	}
}

void mimimi_stage_tick(struct mimimi_stage *stage, struct mimimi_input *input)
{
	struct mimimi_sprite *sprite, *player;
	int i, j, k;
	int x, y;
	int count;
	struct mimimi_image *image;
	struct mimimi_played_sound *sound;
	
	image = &input->image;
	player = stage->sprites;
	
	j = 0;
	while (player->position.x < stage->areas[j].offset.x * 128) j++;
	stage->area_index = j;
	
	for (i = 0 ; i < stage->sprite_count ; i++) {
		
		sprite = stage->sprites + i;
		
		k = 0;
		while (sprite->position.x < stage->areas[k].offset.x * 128) k++;
		sprite->ground = stage->areas[k].ground;
		sprite->offset = stage->areas[k].offset;
		sprite->offset.x *= 128;
		sprite->offset.y *= 128;
		
		stage->clamped_camera.offset = stage->areas[k].offset;
		stage->clamped_camera.size.width = stage->areas[k].ground->width;
		stage->clamped_camera.size.height = stage->areas[k].ground->height;
		
		mimimi_sprite_tick(sprite);
	}
	
	for (i = 0 ; i < stage->platform_set_count ; i++) mimimi_platform_set_tick(stage->platform_sets + i);
	
	x = player->physics.dx / 42;
	y = player->physics.dy / 64 - 256;
	mimimi_camera_tick(&stage->camera, &player->position, x, y);
	mimimi_clamped_camera_tick(&stage->clamped_camera, image->width, image->height);
	
	for (i = 0 ; i < stage->areas[j].background_count ; i++) {
		mimimi_background_tick(stage->areas[j].backgrounds + i, image, &stage->clamped_camera.output, &player->offset);
	}
	for (i = 0 ; i < stage->platform_set_count ; i++) {
		mimimi_platform_set_display_tick(stage->platform_sets + i, image, &stage->clamped_camera.output);
	}
	for (i = 0 ; i < stage->sprite_count ; i++) {
		mimimi_display_tick(image, stage->sprites + i, &stage->clamped_camera.output);
	}
	for (i = 0 ; i < stage->areas[j].foreground_count ; i++) {
		mimimi_background_tick(stage->areas[j].foregrounds + i, image, &stage->clamped_camera.output, &player->offset);
	}
	
	mimimi_boxing(&stage->clamped_camera.size, image);
	
	count = sizeof stage->sounds / sizeof *stage->sounds;
	k = sizeof input->notes / sizeof *input->notes;
	
	for (i = 0 ; i < k ; i++) input->notes[i].volume = 0;
	
	for (i = 0 ; i < count ; i++) {
		sound = stage->sounds + i;
		if (sound->current >= sound->sound.count) continue;
		for (k = 0 ; k < 16 ; k++) input->notes[i * 16 + k] = sound->sound.notes[sound->current][k];
		sound->current++;
	}
}

struct mimimi_sprite *mimimi_spawn(struct mimimi_stage *stage, struct mimimi_model *model, int x, int y, int width, int height)
{
	int i;
	
	if (stage->sprite_count == 0) stage->sprites = NULL;
	
	i = stage->sprite_count++;
	
	stage->sprites = realloc(stage->sprites, sizeof *stage->sprites * stage->sprite_count);
	if (stage->sprites == NULL) exit(1);
	
	mimimi_sprite(stage->sprites + i, model, NULL, x, y, width, height);
	
	if (i == 0) {
		stage->camera = stage->sprites[i].position;
		stage->camera.y -= 256;
	}
	
	return stage->sprites + i;
}

void mimimi_background(struct mimimi_image *image, struct mimimi_ground *ground)
{
	int x, y;
	unsigned char color;
	
	image->width = ground->width * 16;
	image->height = ground->height * 16;
	image->colors = malloc(image->width * image->height);
	if (image->colors == NULL) exit(1);
	
	for (y = 0 ; y < image->height ; y++) {
		for (x = 0 ; x < image->width ; x++) {
			color = ground->colors[mimimi_ground_tile(ground, x / 16, y / 16)];
			image->colors[x + y * image->width] = color;
		}
	}
}

void mimimi_play_sound(struct mimimi_stage *stage, struct mimimi_sound *sound)
{
	int i;
	int count;
	
	count = sizeof stage->sounds / sizeof *stage->sounds;
	
	for (i = 0 ; i < count ; i++) {
		if (stage->sounds[i].current >= stage->sounds[i].sound.count) break;
	}
	
	if (i == count) return;
	
	stage->sounds[i].sound = *sound;
	stage->sounds[i].current = 0;
}

void mimimi_stage_jump(void *data)
{
	static struct mimimi_sound sound = {{0}, 0};
	
	struct mimimi_stage_jump *jump;
	
	jump = data;
	if (jump->sprite->physics.airborne == 0) mimimi_play_sound(jump->stage, &sound);
	mimimi_jump(jump->sprite);
}
