/* license: AGPLv3 or later */
/* copyright 2024 zamfofex */

#ifndef MIMIMI
#define MIMIMI

struct mimimi_position {
	int x, y;
};

struct mimimi_size {
	int width, height;
};

struct mimimi_image {
	int width, height;
	unsigned char *colors;
};

struct mimimi_physics {
	int dx, dy;
	unsigned char airborne;
	int width, height;
	int gravity;
	struct mimimi_sprite *sprite;
};

struct mimimi_walk {
	int ground_speed;
	int airborne_speed;
	int direction;
};

struct mimimi_ground {
	int width, height;
	unsigned int colors[256];
	unsigned int restitution[256];
	unsigned char tiles[0x2000];
};

struct mimimi_sprite {
	struct mimimi_position position;
	struct mimimi_physics physics;
	struct mimimi_walk walk;
	struct mimimi_ground *ground;
	struct mimimi_position offset;
	struct mimimi_model *model;
	unsigned char animation_time;
	unsigned char landing_time;
	int direction;
	int trailing_y;
	unsigned char trail;
};

struct mimimi_positioned_physics {
	struct mimimi_physics *physics;
	struct mimimi_physics other;
	struct mimimi_position transform;
	struct mimimi_position *position;
	struct mimimi_position previous;
	unsigned char was_airborne;
};

struct mimimi_font {
	unsigned char glyphs[0x60][0x10];
};

struct mimimi_pose_layer {
	int slant;
	int y_angle;
	int z_angle;
	int y;
};

struct mimimi_pose {
	int x, y;
	struct mimimi_pose_layer layers[0x80];
};

struct mimimi_background {
	struct mimimi_image image;
	int x, y, z;
};

struct mimimi_row {
	int width;
	int height;
	int count;
	unsigned char colors[0x80];
};

struct mimimi_layer {
	int x, y, z;
	int width;
	int amplifier;
	int count;
	struct mimimi_row rows[0x80];
};

struct mimimi_model_layer {
	int parent_index;
	struct mimimi_layer *layer;
};

struct mimimi_model {
	int count;
	struct mimimi_model_layer layers[0x80];
};

struct mimimi_history {
	unsigned char left, right;
};

struct mimimi_clamped_camera {
	struct mimimi_position *input;
	struct mimimi_position output;
	struct mimimi_position offset;
	struct mimimi_size size;
};

struct mimimi_dialogue_paragraph {
	struct mimimi_model *model;
	char *name;
	char *text;
};

struct mimimi_dialogue {
	struct mimimi_dialogue_paragraph *paragraphs;
	int count;
};

struct mimimi_toast_dialogue {
	int count;
	int *delays;
	int *heights;
	struct mimimi_image *images;
	int i, j, k;
};

struct mimimi_momentum {
	int x, dx;
};

struct mimimi_platform {
	struct mimimi_positioned_physics physics;
	struct mimimi_ground *ground;
	struct mimimi_image *image;
};

struct mimimi_linear_motion {
	int *value;
	int origin;
	int scale;
};

struct mimimi_angular_motion {
	struct mimimi_position *value;
	struct mimimi_position origin;
	int angle;
	int scale;
	int height;
};

struct mimimi_platform_set {
	
	struct mimimi_momentum momentum;
	int balance;
	int strength;
	int linear_weight;
	int angular_weight;
	int min, max;
	
	int linear_platform_count;
	struct mimimi_platform *linear_platforms;
	struct mimimi_linear_motion *linear_motions;
	
	int angular_platform_count;
	struct mimimi_platform *angular_platforms;
	struct mimimi_angular_motion *angular_motions;
};

struct mimimi_area {
	struct mimimi_ground *ground;
	struct mimimi_position offset;
	struct mimimi_background *backgrounds;
	int background_count;
	struct mimimi_background *foregrounds;
	int foreground_count;
};

struct mimimi_note {
	unsigned char frequency, volume;
};

struct mimimi_sound {
	struct mimimi_note notes[10000][16];
	int count;
};

struct mimimi_played_sound {
	struct mimimi_sound sound;
	int current;
};

struct mimimi_stage {
	struct mimimi_sprite *sprites;
	int sprite_count;
	struct mimimi_position camera;
	struct mimimi_clamped_camera clamped_camera;
	struct mimimi_area *areas;
	int area_count, area_index;
	struct mimimi_platform_set *platform_sets;
	int platform_set_count;
	struct mimimi_played_sound sounds[16];
};

struct mimimi_stage_jump {
	struct mimimi_stage *stage;
	struct mimimi_sprite *sprite;
};

struct mimimi_input {
	struct mimimi_image image;
	struct mimimi_note notes[256];
	unsigned char left, right;
};

extern struct mimimi_font *mimimi_font;

extern struct mimimi_model *mimimi_mango;
extern struct mimimi_model *mimimi_pepper;

extern struct mimimi_model *mimimi_homura_flash;
extern struct mimimi_model *mimimi_mirai_flash;
extern struct mimimi_model *mimimi_haruka_flash;

extern struct mimimi_model *mimimi_ryoubi_flash;
extern struct mimimi_model *mimimi_ryouna_flash;

extern struct mimimi_model *mimimi_mirai_shinomas;

extern signed char *mimimi_sine;
extern signed char *mimimi_cosine;

int mimimi_div(int n, int q);
int mimimi_div_up(int n, int q);
int mimimi_div_down(int n, int q);
int mimimi_mod(int n, int q);
int mimimi_angle(int x, int y);

void mimimi_stamp(struct mimimi_image *target, int x, int y, struct mimimi_image *source);

void mimimi_camera_tick(struct mimimi_position *camera, struct mimimi_position *position, int ox, int oy);
void mimimi_camera_stamp(struct mimimi_image *target, struct mimimi_position *camera, int x, int y, int z, struct mimimi_image *source);
void mimimi_clamped_camera_tick(struct mimimi_clamped_camera *camera, int width, int height);

void mimimi_collision_physics_tick(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground);
void mimimi_physics_tick(struct mimimi_physics *physics, struct mimimi_position *position, struct mimimi_ground *ground);
void mimimi_physics(struct mimimi_physics *physics, int width, int height);

void mimimi_sprite(struct mimimi_sprite *sprite, struct mimimi_model *model, struct mimimi_ground *ground, int x, int y, int width, int height);
void mimimi_sprite_tick(struct mimimi_sprite *sprite);

void mimimi_positioned_physics_tick(struct mimimi_positioned_physics *physics, struct mimimi_ground *ground);
void mimimi_positioned_physics(struct mimimi_positioned_physics *physics, struct mimimi_physics *other, struct mimimi_position *position);

int mimimi_measure_character(struct mimimi_font *font, char ch);
int mimimi_draw_character(struct mimimi_font *font, char ch, struct mimimi_image *target, int x, int y, unsigned char color);

int mimimi_measure_segment(struct mimimi_font *font, char *text, int count);
int mimimi_draw_segment(struct mimimi_font *font, char *text, int count, struct mimimi_image *target, int x, int y, unsigned char color);

char *mimimi_skip_word(char *text);
int mimimi_count_word(char *text);
int mimimi_measure_word(struct mimimi_font *font, char *text);
int mimimi_draw_word(struct mimimi_font *font, char *text, struct mimimi_image *target, int x, int y, unsigned char color);

char *mimimi_skip_paragraph(char *text);
int mimimi_count_paragraph(char *text);
int mimimi_measure_paragraph(struct mimimi_font *font, char *text);
int mimimi_draw_paragraph(struct mimimi_font *font, char *text, struct mimimi_image *target, int x, int y, unsigned char color);

char *mimimi_skip_line(struct mimimi_font *font, char *text, int width);
int mimimi_count_line(struct mimimi_font *font, char *text, int width);
int mimimi_measure_line(struct mimimi_font *font, char *text, int width);
int mimimi_draw_line(struct mimimi_font *font, char *text, int width, struct mimimi_image *target, int x, int y, unsigned char color);

int mimimi_measure_text(struct mimimi_font *font, char *text, int width, int line_height);
int mimimi_draw_text(struct mimimi_font *font, char *text, int width, int line_height, struct mimimi_image *target, int x, int y, unsigned char color);

void mimimi_pose(struct mimimi_image *image, struct mimimi_model *model, struct mimimi_pose *pose);
void mimimi_interpolate(struct mimimi_pose *a, struct mimimi_pose *b, int i, int count);

void mimimi_background_tick(struct mimimi_background *background, struct mimimi_image *image, struct mimimi_position *camera, struct mimimi_position *offset);
void mimimi_display_tick(struct mimimi_image *image, struct mimimi_sprite *sprite, struct mimimi_position *camera);

unsigned char mimimi_ground_tile(struct mimimi_ground *ground, int x, int y);

void mimimi_history_tick(struct mimimi_history *history, unsigned char left, unsigned char right);
void mimimi_controls_tick(struct mimimi_walk *walk, struct mimimi_history *history, void (*jump)(void *data), void *data);
void mimimi_jump(void *data);

void mimimi_toast_dialogue(struct mimimi_toast_dialogue *toast_dialogue, struct mimimi_dialogue *dialogue);
int mimimi_toast_dialogue_tick(struct mimimi_toast_dialogue *dialogue, struct mimimi_image *image);

void mimimi_stage(struct mimimi_stage *stage, struct mimimi_ground **grounds, int count);
void mimimi_stage_tick(struct mimimi_stage *stage, struct mimimi_input *input);
struct mimimi_sprite *mimimi_spawn(struct mimimi_stage *stage, struct mimimi_model *model, int x, int y, int width, int height);
void mimimi_stage_jump(void *data);
void mimimi_play_sound(struct mimimi_stage *stage, struct mimimi_sound *sound);

void mimimi_linear_propagate(struct mimimi_momentum *momentum, int value, int x0, int scale);
void mimimi_linear_apply(struct mimimi_momentum *momentum, int *value, int x0, int scale);

void mimimi_angular_propagate(struct mimimi_momentum *momentum, struct mimimi_position *position, int x0, int y0, int a0, int scale);
void mimimi_angular_apply(struct mimimi_momentum *momentum, struct mimimi_position *position, int x0, int y0, int a0, int scale, int length);

void mimimi_linear_motion_propagate(struct mimimi_linear_motion *motion, struct mimimi_momentum *momentum);
void mimimi_linear_motion_apply(struct mimimi_linear_motion *motion, struct mimimi_momentum *momentum);

void mimimi_angular_motion_propagate(struct mimimi_angular_motion *motion, struct mimimi_momentum *momentum);
void mimimi_angular_motion_apply(struct mimimi_angular_motion *motion, struct mimimi_momentum *momentum);

void mimimi_platform_display_tick(struct mimimi_platform *platform, struct mimimi_position *camera, struct mimimi_image *image);
struct mimimi_platform_set *mimimi_add_platform_set(struct mimimi_stage *stage);
void mimimi_spawn_linear_platform(struct mimimi_stage *stage, struct mimimi_platform_set *platforms, struct mimimi_ground *ground, struct mimimi_image *image, int x, int y, int scale);
void mimimi_spawn_angular_platform(struct mimimi_stage *stage, struct mimimi_platform_set *platforms, struct mimimi_ground *ground, struct mimimi_image *image, int x, int y, int a, int scale, int height);

void mimimi_platform_set_tick(struct mimimi_platform_set *platform_set);
void mimimi_platform_set_display_tick(struct mimimi_platform_set *platforms, struct mimimi_image *image, struct mimimi_position *camera);

void mimimi_unpack_bits(int size, unsigned char *unpacked, unsigned char *packed);

void mimimi_linear_motion(struct mimimi_linear_motion *motion, int *value, int origin, int scale);
void mimimi_angular_motion(struct mimimi_angular_motion *motion, struct mimimi_position *value, int x0, int y0, int a0, int scale, int height);

void mimimi_background(struct mimimi_image *image, struct mimimi_ground *ground);

void mimimi_boxing(struct mimimi_size *size, struct mimimi_image *image);

void mimimi_scale(struct mimimi_image *target, struct mimimi_image *source);

void mimimi_main(void (*tick)(void *data, struct mimimi_input *input), void *data);

#endif
